@@ -304,7 +304,7 @@ void *dtlsservernew(void *arg) {
     unsigned long error;
     struct timeval timeout;
     struct addrinfo tmpsrvaddr;
-    char tmp[INET6_ADDRSTRLEN];
+    char tmp[INET6_ADDRSTRLEN], *subj;
 
     debug(DBG_WARN, "dtlsservernew: incoming DTLS connection from %s", addr2string((struct sockaddr *)&params->addr, tmp, sizeof(tmp)));
 
@@ -343,6 +343,12 @@ void *dtlsservernew(void *arg) {
 
     while (conf) {
         if (accepted_tls == conf->tlsconf && verifyconfcert(cert, conf)) {
+            subj = getcertsubject(cert);
+            if(subj) {
+                debug(DBG_WARN, "dtlsservernew: DTLS connection from %s, client %s, subject %s up",
+                    addr2string((struct sockaddr *)&params->addr, tmp, sizeof(tmp)), conf->name, subj);
+                free(subj);
+            }
             X509_free(cert);
             client = addclient(conf, 1);
             if (client) {
@@ -419,7 +425,7 @@ int getConnectionInfo(int socket, struct sockaddr *from, socklen_t fromlen, stru
 #endif
         if(ctrlhdr->cmsg_level == IPPROTO_IPV6 && ctrlhdr->cmsg_type == IPV6_RECVPKTINFO) {
             info6 = (struct in6_pktinfo *)CMSG_DATA(ctrlhdr);
-            debug(DBG_DBG, "udp packet to: %x", inet_ntop(AF_INET6, &info6->ipi6_addr, tmp, sizeof(tmp)));
+            debug(DBG_DBG, "udp packet to: %s", inet_ntop(AF_INET6, &info6->ipi6_addr, tmp, sizeof(tmp)));
 
             ((struct sockaddr_in6 *)to)->sin6_addr = info6->ipi6_addr;
             ((struct sockaddr_in6 *)to)->sin6_scope_id = info6->ipi6_ifindex;
@@ -523,6 +529,8 @@ int dtlsconnect(struct server *server, int timeout, char *text) {
     struct hostportres *hp;
     unsigned long error;
     BIO *bio;
+    struct addrinfo *source = NULL;
+    char *subj;
 
     debug(DBG_DBG, "dtlsconnect: called from %s", text);
     pthread_mutex_lock(&server->lock);
@@ -533,6 +541,13 @@ int dtlsconnect(struct server *server, int timeout, char *text) {
     pthread_mutex_unlock(&server->lock);
 
     hp = (struct hostportres *)list_first(server->conf->hostports)->data;
+
+    if(server->conf->source) {
+        source = resolvepassiveaddrinfo(server->conf->source, AF_UNSPEC, NULL, protodefs.socktype);
+        if(!source)
+            debug(DBG_WARN, "dtlsconnect: could not resolve source address to bind for server %s, using default", server->conf->name);
+    }
+
     gettimeofday(&start, NULL);
 
     for (;;) {
@@ -546,19 +561,19 @@ int dtlsconnect(struct server *server, int timeout, char *text) {
         server->ssl = NULL;
 
         wait = connect_wait(start, server->connecttime, firsttry);
+        gettimeofday(&now, NULL);
+        if (timeout && (now.tv_sec - start.tv_sec) + wait > timeout) {
+            debug(DBG_DBG, "dtlsconnect: timeout");
+            if (source) freeaddrinfo(source);
+            return 0;
+        }
         debug(DBG_INFO, "Next connection attempt to %s in %lds", server->conf->name, wait);
         sleep(wait);
         firsttry = 0;
 
-        gettimeofday(&now, NULL);
-        if (timeout && (now.tv_sec - start.tv_sec) > timeout) {
-            debug(DBG_DBG, "tlsconnect: timeout");
-            return 0;
-        }
-
         debug(DBG_INFO, "dtlsconnect: connecting to %s port %s", hp->host, hp->port);
 
-        if ((server->sock = bindtoaddr(srcres, hp->addrinfo->ai_family, 0)) < 0)
+        if ((server->sock = bindtoaddr(source ? source : srcres, hp->addrinfo->ai_family, 0)) < 0)
             continue;
         if (connect(server->sock, hp->addrinfo->ai_addr, hp->addrinfo->ai_addrlen))
             continue;
@@ -580,7 +595,7 @@ int dtlsconnect(struct server *server, int timeout, char *text) {
         if (sslconnecttimeout(server->ssl, 5) <= 0) {
             while ((error = ERR_get_error()))
                 debug(DBG_ERR, "dtlsconnect: SSL connect to %s failed: %s", server->conf->name, ERR_error_string(error, NULL));
-            debug(DBG_ERR, "dtlsconnect: SSL connect to %s failed", server->conf->name, ERR_error_string(error, NULL));
+            debug(DBG_ERR, "dtlsconnect: SSL connect to %s failed", server->conf->name);
             continue;
         }
         socktimeout.tv_sec = 5;
@@ -592,12 +607,16 @@ int dtlsconnect(struct server *server, int timeout, char *text) {
         if (!cert)
             continue;
         if (verifyconfcert(cert, server->conf)) {
+            subj = getcertsubject(cert);
+            if(subj) {
+                debug(DBG_WARN, "dtlsconnect: DTLS connection to %s, subject %s up", server->conf->name, subj);
+                free(subj);
+            }
             X509_free(cert);
             break;
         }
         X509_free(cert);
     }
-    debug(DBG_WARN, "dtlsconnect: DTLS connection to %s up", server->conf->name);
 
     pthread_mutex_lock(&server->lock);
     server->state = RSP_SERVER_STATE_CONNECTED;
@@ -607,6 +626,7 @@ int dtlsconnect(struct server *server, int timeout, char *text) {
     server->conreset = 1;
     pthread_cond_signal(&server->newrq_cond);
     pthread_mutex_unlock(&server->newrq_mutex);
+    if (source) freeaddrinfo(source);
     return 1;
 }
 
@@ -635,7 +655,7 @@ int clientradputdtls(struct server *server, unsigned char *rad) {
                 return 0;
         }
     }
-    debug(DBG_DBG, "clientradputdtls: Sent %d bytes, Radius packet of length %d to DTLS peer %s", cnt, len, conf->name);
+    debug(DBG_DBG, "clientradputdtls: Sent %d bytes, Radius packet of length %zu to DTLS peer %s", cnt, len, conf->name);
     pthread_mutex_unlock(&server->lock);
     return 1;
 }
