@@ -95,7 +95,7 @@ struct hostportres *newhostport(char *hostport, char *default_port, uint8_t pref
     slash = hostport ? strchr(hostport, '/') : NULL;
     if (slash) {
 	if (!prefixok) {
-	    debug(DBG_WARN, "newhostport: prefix not allowed here", hp->host);
+	    debug(DBG_WARN, "newhostport: prefix not allowed here (%s)", hp->host);
 	    goto errexit;
 	}
 	s = slash + 1;
@@ -221,14 +221,28 @@ int resolvehostports(struct list *hostports, int af, int socktype) {
     return 1;
 }
 
-struct addrinfo *resolvepassiveaddrinfo(char *hostport, int af, char *default_port, int socktype) {
-    struct addrinfo *ai = NULL;
-    struct hostportres *hp = newhostport(hostport, default_port, 0);
-    if (hp && resolvehostport(hp, af, socktype, 1)) {
-	ai = hp->addrinfo;
-	hp->addrinfo = NULL;
+struct addrinfo *resolvepassiveaddrinfo(char **hostport, int af, char *default_port, int socktype) {
+    struct addrinfo *ai = NULL, *last_ai;
+    int i;
+    char *any[2] = {"*", NULL};
+
+    if(!hostport)
+        hostport = any;
+
+    for (i = 0; hostport[i]; i++) {
+        struct hostportres *hp = newhostport(hostport[i], default_port, 0);
+        if (hp && resolvehostport(hp, af, socktype, 1)) {
+            if (!ai) {
+                ai = last_ai = hp->addrinfo;
+            } else {
+                last_ai->ai_next = hp->addrinfo;
+            }
+            while (last_ai->ai_next)
+                last_ai = last_ai->ai_next;
+            hp->addrinfo = NULL;
+        }
+        freehostport(hp);
     }
-    freehostport(hp);
     return ai;
 }
 
