@@ -15,7 +15,7 @@ void freehostport(struct hostportres *hp);
 void freehostports(struct list *hostports);
 int resolvehostport(struct hostportres *hp, int af, int socktype, uint8_t passive);
 int resolvehostports(struct list *hostports, int af, int socktype);
-struct addrinfo *resolvepassiveaddrinfo(char *hostport, int af, char *default_port, int socktype);
+struct addrinfo *resolvepassiveaddrinfo(char **hostport, int af, char *default_port, int socktype);
 int hostportmatches(struct list *hostports, struct list *matchhostports, uint8_t checkport);
 int addressmatches(struct list *hostports, struct sockaddr *addr, uint8_t checkport);
 int connecttcphostlist(struct list *hostports,  struct addrinfo *src);
