@@ -56,6 +56,7 @@
 #include <pthread.h>
 #include <errno.h>
 #include <assert.h>
+#include <poll.h>
 #include <openssl/ssl.h>
 #include <openssl/rand.h>
 #include <openssl/err.h>
@@ -841,14 +842,16 @@ int msmppe(unsigned char *attrs, int length, uint8_t type, char *attrtxt, struct
 
 int rewriteusername(struct request *rq, struct tlv *attr) {
     char *orig = (char *)tlv2str(attr);
+    if (!orig)
+        return 0;
     if (!dorewritemodattr(attr, rq->from->conf->rewriteusername)) {
-	free(orig);
-	return 0;
+        free(orig);
+        return 0;
     }
     if (strlen(orig) != attr->l || memcmp(orig, attr->v, attr->l))
-	rq->origusername = (char *)orig;
+        rq->origusername = (char *)orig;
     else
-	free(orig);
+        free(orig);
     return 1;
 }
 
@@ -978,6 +981,7 @@ uint8_t *radattr2ascii(struct tlv *attr) {
 
 void replylog(struct radmsg *msg, struct server *server, struct request *rq) {
     uint8_t *username, *logusername = NULL, *stationid, *replymsg, *tmpmsg;
+    uint8_t *operatorname, *cui;
     char *servername, *logstationid = NULL;
     uint8_t level = DBG_NOTICE;
     char tmp[INET6_ADDRSTRLEN];
@@ -1012,6 +1016,20 @@ void replylog(struct radmsg *msg, struct server *server, struct request *rq) {
         }
         free(stationid);
     }
+    cui = radattr2ascii(radmsg_gettype(msg, RAD_Attr_CUI));
+    if(cui) {
+        if (asprintf((char **)&tmpmsg, " cui %s", cui) >= 0) {
+            free(cui);
+            cui = tmpmsg;
+        }
+    }
+    operatorname = radattr2ascii(radmsg_gettype(rq->msg, RAD_Attr_Operator_Name));
+    if (operatorname) {
+        if (asprintf((char **)&tmpmsg, " operator %s", operatorname) >= 0) {
+            free(operatorname);
+            operatorname = tmpmsg;
+        }
+    }
     replymsg = radattr2ascii(radmsg_gettype(msg, RAD_Attr_Reply_Message));
     if (replymsg) {
         if (asprintf((char **)&tmpmsg, " (%s)", replymsg) >= 0) {
@@ -1024,10 +1042,10 @@ void replylog(struct radmsg *msg, struct server *server, struct request *rq) {
         if (msg->code == RAD_Accounting_Response)
             level = DBG_INFO;
         if (logusername) {
-            debug(level, "%s for user %s%s from %s%s to %s (%s)",
-                radmsgtype2string(msg->code), logusername, logstationid ? logstationid : "",
+            debug(level, "%s for user %s%s%s from %s%s to %s (%s)%s",
+                radmsgtype2string(msg->code), logusername, logstationid ? logstationid : "", cui ? (char *)cui : "",
                 servername, replymsg ? (char *)replymsg : "", rq->from->conf->name,
-                addr2string(rq->from->addr, tmp, sizeof(tmp)));
+                addr2string(rq->from->addr, tmp, sizeof(tmp)), operatorname ? (char *)operatorname : "");
         } else {
             debug(level, "%s (response to %s) from %s to %s (%s)", radmsgtype2string(msg->code),
                 radmsgtype2string(rq->msg->code), servername,
@@ -1040,6 +1058,8 @@ void replylog(struct radmsg *msg, struct server *server, struct request *rq) {
     }
     free(username);
     free(logstationid);
+    free(cui);
+    free(operatorname);
     free(replymsg);
 }
 
@@ -1567,23 +1587,22 @@ void *clientwr(void *arg) {
 
 #define ZZZ 900
 
-    server->state = RSP_SERVER_STATE_STARTUP;
+    if (server->state != RSP_SERVER_STATE_BLOCKING_STARTUP)
+        server->state = RSP_SERVER_STATE_STARTUP;
     if (server->dynamiclookuparg && !dynamicconfig(server)) {
-	dynconffail = 1;
-	server->state = RSP_SERVER_STATE_FAILING;
-	debug(DBG_WARN, "%s: dynamicconfig(%s: %s) failed, sleeping %ds",
+        dynconffail = 1;
+        server->state = RSP_SERVER_STATE_FAILING;
+        debug(DBG_WARN, "%s: dynamicconfig(%s: %s) failed, sleeping %ds",
               __func__, server->conf->name, server->dynamiclookuparg, ZZZ);
-	sleep(ZZZ);
-	goto errexit;
+        goto errexitwait;
     }
     /* FIXME: Is resolving not always done by compileserverconfig(),
      * either as part of static configuration setup or by
      * dynamicconfig() above?  */
     if (!resolvehostports(conf->hostports, conf->hostaf, conf->pdef->socktype)) {
         debug(DBG_WARN, "%s: resolve failed, sleeping %ds", __func__, ZZZ);
-	server->state = RSP_SERVER_STATE_FAILING;
-        sleep(ZZZ);
-        goto errexit;
+        server->state = RSP_SERVER_STATE_FAILING;
+        goto errexitwait;
     }
 
     memset(&timeout, 0, sizeof(struct timespec));
@@ -1593,20 +1612,19 @@ void *clientwr(void *arg) {
     laststatsrv = server->lastreply;
 
     if (conf->pdef->connecter) {
-	if (!conf->pdef->connecter(server, server->dynamiclookuparg ? 5 : 0, "clientwr")) {
-	    server->state = RSP_SERVER_STATE_FAILING;
-	    if (server->dynamiclookuparg) {
-                debug(DBG_WARN, "%s: connect failed, sleeping %ds",
-                      __func__, ZZZ);
-		sleep(ZZZ);
-	    }
-	    goto errexit;
-	}
-	if (pthread_create(&clientrdth, &pthread_attr, conf->pdef->clientconnreader, (void *)server)) {
-	    debugerrno(errno, DBG_ERR, "clientwr: pthread_create failed");
-	    server->state = RSP_SERVER_STATE_FAILING;
-	    goto errexit;
-	}
+        if (!conf->pdef->connecter(server, server->dynamiclookuparg ? 5 : 0, "clientwr")) {
+            server->state = RSP_SERVER_STATE_FAILING;
+            if (server->dynamiclookuparg) {
+                debug(DBG_WARN, "%s: connect failed, sleeping %ds", __func__, ZZZ);
+                goto errexitwait;
+            }
+            goto errexit;
+        }
+        if (pthread_create(&clientrdth, &pthread_attr, conf->pdef->clientconnreader, (void *)server)) {
+            debugerrno(errno, DBG_ERR, "clientwr: pthread_create failed");
+            server->state = RSP_SERVER_STATE_FAILING;
+            goto errexit;
+        }
     }
     server->state = RSP_SERVER_STATE_CONNECTED;
 
@@ -1741,6 +1759,16 @@ void *clientwr(void *arg) {
         }
     }
     }
+
+errexitwait:
+    /* flush request queue so we don't block incoming retries by removing blocked duplicates*/
+    for (i=0; i < MAX_REQUESTS; i++) {
+        rqout = server->requests + i;
+        pthread_mutex_lock(rqout->lock);
+        freerqoutdata(rqout);
+        pthread_mutex_unlock(rqout->lock);
+    }
+    sleep(ZZZ);
 errexit:
     if (server->dynamiclookuparg) {
 	removeserversubrealms(realms, conf);
@@ -2040,7 +2068,7 @@ struct list *createsubrealmservers(struct realm *realm, struct list *srvconfs) {
              * the srvconfs list.  */
 	    if (addserver(srvconf)) {
 		srvconf->servers->dynamiclookuparg = stringcopy(realm->name, 0);
-		srvconf->servers->state = RSP_SERVER_STATE_STARTUP;
+		srvconf->servers->state = srvconf->blockingstartup ? RSP_SERVER_STATE_BLOCKING_STARTUP : RSP_SERVER_STATE_STARTUP;
                 debug(DBG_DBG, "%s: new client writer for %s",
                       __func__, srvconf->servers->conf->name);
 		if (pthread_create(&clientth, &pthread_attr, clientwr, (void *)(srvconf->servers))) {
@@ -2098,10 +2126,12 @@ struct realm *adddynamicrealmserver(struct realm *realm, char *id) {
 }
 
 int dynamicconfig(struct server *server) {
-    int ok, fd[2], status;
+    int ok = 0, fd[2], status;
     pid_t pid;
     struct clsrvconf *conf = server->conf;
     struct gconffile *cf = NULL;
+    struct pollfd fds[1];
+    FILE *pipein;
 
     /* for now we only learn hostname/address */
     debug(DBG_DBG, "dynamicconfig: need dynamic server config for %s", server->dynamiclookuparg);
@@ -2129,10 +2159,25 @@ int dynamicconfig(struct server *server) {
     }
 
     close(fd[1]);
-    pushgconffile(&cf, fdopen(fd[0], "r"), conf->dynamiclookupcommand);
-    ok = getgenericconfig(&cf, NULL, "Server", CONF_CBK, confserver_cb,
-			  (void *) conf, NULL);
-    freegconf(&cf);
+    pipein = fdopen(fd[0], "r");
+    fds[0].fd = fd[0];
+    fds[0].events = POLLIN;
+    status = poll(fds, 1, 5000);
+    if (status < 0) {
+        debugerrno(errno, DBG_ERR, "dynamicconfig: error while waiting for command output");
+    } else if (status ==0) {
+        debug(DBG_WARN, "dynamicconfig: command did not return anything in time");
+        kill(pid, SIGKILL);
+    } else {
+        pipein = pushgconffile(&cf, pipein, conf->dynamiclookupcommand);
+        if (pipein) {
+            ok = getgenericconfig(&cf, NULL, "Server", CONF_CBK, confserver_cb, (void *) conf, NULL);
+            freegconf(&cf);
+            pipein = NULL;
+        }
+    }
+    if (pipein)
+        fclose(pipein);
 
     if (waitpid(pid, &status, 0) < 0) {
 	debugerrno(errno, DBG_ERR, "dynamicconfig: wait error");
@@ -2170,14 +2215,12 @@ void freeclsrvconf(struct clsrvconf *conf) {
     if (conf->hostsrc)
 	freegconfmstr(conf->hostsrc);
     free(conf->portsrc);
+    freegconfmstr(conf->source);
     free(conf->confsecret);
     free(conf->secret);
     free(conf->tls);
-    free(conf->matchcertattr);
-    if (conf->certcnregex)
-	regfree(conf->certcnregex);
-    if (conf->certuriregex)
-	regfree(conf->certuriregex);
+    freegconfmstr(conf->confmatchcertattrs);
+    freematchcertattr(conf);
     free(conf->confrewritein);
     free(conf->confrewriteout);
     if (conf->rewriteusername) {
@@ -2264,9 +2307,10 @@ int mergesrvconf(struct clsrvconf *dst, struct clsrvconf *src) {
     if (!mergeconfstring(&dst->name, &src->name) ||
 	!mergeconfmstring(&dst->hostsrc, &src->hostsrc) ||
 	!mergeconfstring(&dst->portsrc, &src->portsrc) ||
+    !mergeconfmstring(&dst->source, &src->source) ||
 	!mergeconfstring(&dst->confsecret, &src->confsecret) ||
 	!mergeconfstring(&dst->tls, &src->tls) ||
-	!mergeconfstring(&dst->matchcertattr, &src->matchcertattr) ||
+	!mergeconfmstring(&dst->confmatchcertattrs, &src->confmatchcertattrs) ||
 	!mergeconfstring(&dst->confrewritein, &src->confrewritein) ||
 	!mergeconfstring(&dst->confrewriteout, &src->confrewriteout) ||
 	!mergeconfstring(&dst->confrewriteusername, &src->confrewriteusername) ||
@@ -2282,6 +2326,7 @@ int mergesrvconf(struct clsrvconf *dst, struct clsrvconf *src) {
 	dst->retryinterval = src->retryinterval;
     if (src->retrycount != 255)
 	dst->retrycount = src->retrycount;
+    dst->blockingstartup = src->blockingstartup;
     return 1;
 }
 
@@ -2307,10 +2352,11 @@ int config_hostaf(const char *desc, int ipv4only, int ipv6only, int *af) {
 
 int confclient_cb(struct gconffile **cf, void *arg, char *block, char *opt, char *val) {
     struct clsrvconf *conf, *existing;
-    char *conftype = NULL, *rewriteinalias = NULL;
+    char *conftype = NULL, *rewriteinalias = NULL, **matchcertattrs = NULL;
     long int dupinterval = LONG_MIN, addttl = LONG_MIN;
     uint8_t ipv4only = 0, ipv6only = 0;
     struct list_node *entry;
+    int i;
 
     debug(DBG_DBG, "confclient_cb called for %s", block);
 
@@ -2329,7 +2375,7 @@ int confclient_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
 	    "secret", CONF_STR_NOESC, &conf->confsecret,
 #if defined(RADPROT_TLS) || defined(RADPROT_DTLS)
 	    "tls", CONF_STR, &conf->tls,
-	    "matchcertificateattribute", CONF_STR, &conf->matchcertattr,
+	    "MatchCertificateAttribute", CONF_MSTR, &matchcertattrs,
 	    "CertificateNameCheck", CONF_BLN, &conf->certnamecheck,
 #endif
 	    "DuplicateInterval", CONF_LINT, &dupinterval,
@@ -2366,13 +2412,19 @@ int confclient_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
 
 #if defined(RADPROT_TLS) || defined(RADPROT_DTLS)
     if (conf->type == RAD_TLS || conf->type == RAD_DTLS) {
-	conf->tlsconf = conf->tls
-            ? tlsgettls(conf->tls, NULL)
-            : tlsgettls("defaultClient", "default");
-	if (!conf->tlsconf)
-	    debugx(1, DBG_ERR, "error in block %s, no tls context defined", block);
-	if (conf->matchcertattr && !addmatchcertattr(conf))
-	    debugx(1, DBG_ERR, "error in block %s, invalid MatchCertificateAttributeValue", block);
+        conf->tlsconf = conf->tls
+                ? tlsgettls(conf->tls, NULL)
+                : tlsgettls("defaultClient", "default");
+        if (!conf->tlsconf)
+            debugx(1, DBG_ERR, "error in block %s, no tls context defined", block);
+        if (matchcertattrs) {
+            for (i=0; matchcertattrs[i]; i++){
+                if (!addmatchcertattr(conf, matchcertattrs[i])) {
+                    debugx(1, DBG_ERR, "error in block %s, invalid MatchCertificateAttributeValue", block);
+                }
+            }
+            freegconfmstr(matchcertattrs);
+        }
     }
 #endif
 
@@ -2449,19 +2501,23 @@ int confclient_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
 }
 
 int compileserverconfig(struct clsrvconf *conf, const char *block) {
+    int i;
 #if defined(RADPROT_TLS) || defined(RADPROT_DTLS)
     if (conf->type == RAD_TLS || conf->type == RAD_DTLS) {
     	conf->tlsconf = conf->tls
             ? tlsgettls(conf->tls, NULL)
             : tlsgettls("defaultServer", "default");
-	if (!conf->tlsconf) {
-	    debug(DBG_ERR, "error in block %s, no tls context defined", block);
-	    return 0;
-	}
-	if (conf->matchcertattr && !addmatchcertattr(conf)) {
-	    debug(DBG_ERR, "error in block %s, invalid MatchCertificateAttributeValue", block);
-	    return 0;
-	}
+        if (!conf->tlsconf) {
+            debug(DBG_ERR, "error in block %s, no tls context defined", block);
+            return 0;
+        }
+        if (conf->confmatchcertattrs) {
+            for (i=0; conf->confmatchcertattrs[i]; i++){
+                if (!addmatchcertattr(conf, conf->confmatchcertattrs[i])) {
+                    debugx(1, DBG_ERR, "error in block %s, invalid MatchCertificateAttributeValue", block);
+                }
+            }
+        }
     }
 #endif
 
@@ -2522,28 +2578,30 @@ int confserver_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
     }
 
     if (!getgenericconfig(cf, block,
-			  "type", CONF_STR, &conftype,
-			  "host", CONF_MSTR, &conf->hostsrc,
+            "type", CONF_STR, &conftype,
+            "host", CONF_MSTR, &conf->hostsrc,
                           "IPv4Only", CONF_BLN, &ipv4only,
                           "IPv6Only", CONF_BLN, &ipv6only,
-			  "port", CONF_STR, &conf->portsrc,
-			  "secret", CONF_STR_NOESC, &conf->confsecret,
+            "port", CONF_STR, &conf->portsrc,
+            "source", CONF_MSTR, &conf->source,
+            "secret", CONF_STR_NOESC, &conf->confsecret,
 #if defined(RADPROT_TLS) || defined(RADPROT_DTLS)
-			  "tls", CONF_STR, &conf->tls,
-			  "MatchCertificateAttribute", CONF_STR, &conf->matchcertattr,
-			  "CertificateNameCheck", CONF_BLN, &conf->certnamecheck,
+            "tls", CONF_STR, &conf->tls,
+            "MatchCertificateAttribute", CONF_MSTR, &conf->confmatchcertattrs,
+            "CertificateNameCheck", CONF_BLN, &conf->certnamecheck,
 #endif
-			  "addTTL", CONF_LINT, &addttl,
-              "tcpKeepalive", CONF_BLN, &conf->keepalive,
-			  "rewrite", CONF_STR, &rewriteinalias,
-			  "rewriteIn", CONF_STR, &conf->confrewritein,
-			  "rewriteOut", CONF_STR, &conf->confrewriteout,
-			  "StatusServer", CONF_STR, &statusserver,
-			  "RetryInterval", CONF_LINT, &retryinterval,
-			  "RetryCount", CONF_LINT, &retrycount,
-			  "DynamicLookupCommand", CONF_STR, &conf->dynamiclookupcommand,
-			  "LoopPrevention", CONF_BLN, &conf->loopprevention,
-			  NULL
+            "addTTL", CONF_LINT, &addttl,
+            "tcpKeepalive", CONF_BLN, &conf->keepalive,
+            "rewrite", CONF_STR, &rewriteinalias,
+            "rewriteIn", CONF_STR, &conf->confrewritein,
+            "rewriteOut", CONF_STR, &conf->confrewriteout,
+            "StatusServer", CONF_STR, &statusserver,
+            "RetryInterval", CONF_LINT, &retryinterval,
+            "RetryCount", CONF_LINT, &retrycount,
+            "DynamicLookupCommand", CONF_STR, &conf->dynamiclookupcommand,
+            "LoopPrevention", CONF_BLN, &conf->loopprevention,
+            "BlockingStartup", CONF_BLN, &conf->blockingstartup,
+            NULL
 	    )) {
 	debug(DBG_ERR, "configuration error");
 	goto errexit;
@@ -2590,7 +2648,7 @@ int confserver_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
 
     if (retryinterval != LONG_MIN) {
 	if (retryinterval < 1 || retryinterval > conf->pdef->retryintervalmax) {
-	    debug(DBG_ERR, "error in block %s, value of option RetryInterval is %d, must be 1-%d", block, retryinterval, conf->pdef->retryintervalmax);
+	    debug(DBG_ERR, "error in block %s, value of option RetryInterval is %ld, must be 1-%d", block, retryinterval, conf->pdef->retryintervalmax);
 	    goto errexit;
 	}
 	conf->retryinterval = (uint8_t)retryinterval;
@@ -2599,7 +2657,7 @@ int confserver_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
 
     if (retrycount != LONG_MIN) {
 	if (retrycount < 0 || retrycount > conf->pdef->retrycountmax) {
-	    debug(DBG_ERR, "error in block %s, value of option RetryCount is %d, must be 0-%d", block, retrycount, conf->pdef->retrycountmax);
+	    debug(DBG_ERR, "error in block %s, value of option RetryCount is %ld, must be 0-%d", block, retrycount, conf->pdef->retrycountmax);
 	    goto errexit;
 	}
 	conf->retrycount = (uint8_t)retrycount;
@@ -2608,7 +2666,7 @@ int confserver_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
 
     if (addttl != LONG_MIN) {
 	if (addttl < 1 || addttl > 255) {
-	    debug(DBG_ERR, "error in block %s, value of option addTTL is %d, must be 1-255", block, addttl);
+	    debug(DBG_ERR, "error in block %s, value of option addTTL is %ld, must be 1-255", block, addttl);
 	    goto errexit;
 	}
 	conf->addttl = (uint8_t)addttl;
@@ -2728,7 +2786,7 @@ int confrealm_cb(struct gconffile **cf, void *arg, char *block, char *opt, char
     return 1;
 }
 
-int setprotoopts(uint8_t type, char **listenargs, char *sourcearg) {
+int setprotoopts(uint8_t type, char **listenargs, char **sourcearg) {
     struct commonprotoopts *protoopts;
 
     protoopts = malloc(sizeof(struct commonprotoopts));
@@ -2745,7 +2803,7 @@ void getmainconfig(const char *configfile) {
     long int addttl = LONG_MIN, loglevel = LONG_MIN;
     struct gconffile *cfs;
     char **listenargs[RAD_PROTOCOUNT];
-    char *sourcearg[RAD_PROTOCOUNT];
+    char **sourceargs[RAD_PROTOCOUNT];
     char *log_mac_str = NULL;
     char *log_key_str = NULL;
     uint8_t *fticks_reporting_str = NULL;
@@ -2756,7 +2814,7 @@ void getmainconfig(const char *configfile) {
     cfs = openconfigfile(configfile);
     memset(&options, 0, sizeof(options));
     memset(&listenargs, 0, sizeof(listenargs));
-    memset(&sourcearg, 0, sizeof(sourcearg));
+    memset(&sourceargs, 0, sizeof(sourceargs));
     options.logfullusername = 1;
 
     clconfs = list_create();
@@ -2775,19 +2833,19 @@ void getmainconfig(const char *configfile) {
 	    &cfs, NULL,
 #ifdef RADPROT_UDP
 	    "ListenUDP", CONF_MSTR, &listenargs[RAD_UDP],
-	    "SourceUDP", CONF_STR, &sourcearg[RAD_UDP],
+	    "SourceUDP", CONF_MSTR, &sourceargs[RAD_UDP],
 #endif
 #ifdef RADPROT_TCP
 	    "ListenTCP", CONF_MSTR, &listenargs[RAD_TCP],
-	    "SourceTCP", CONF_STR, &sourcearg[RAD_TCP],
+	    "SourceTCP", CONF_MSTR, &sourceargs[RAD_TCP],
 #endif
 #ifdef RADPROT_TLS
 	    "ListenTLS", CONF_MSTR, &listenargs[RAD_TLS],
-	    "SourceTLS", CONF_STR, &sourcearg[RAD_TLS],
+	    "SourceTLS", CONF_MSTR, &sourceargs[RAD_TLS],
 #endif
 #ifdef RADPROT_DTLS
 	    "ListenDTLS", CONF_MSTR, &listenargs[RAD_DTLS],
-	    "SourceDTLS", CONF_STR, &sourcearg[RAD_DTLS],
+	    "SourceDTLS", CONF_MSTR, &sourceargs[RAD_DTLS],
 #endif
             "PidFile", CONF_STR, &options.pidfile,
 	    "TTLAttribute", CONF_STR, &options.ttlattr,
@@ -2863,8 +2921,8 @@ void getmainconfig(const char *configfile) {
 		     &fticks_key_str);
 
     for (i = 0; i < RAD_PROTOCOUNT; i++)
-	if (listenargs[i] || sourcearg[i])
-	    setprotoopts(i, listenargs[i], sourcearg[i]);
+	if (listenargs[i] || sourceargs[i])
+	    setprotoopts(i, listenargs[i], sourceargs[i]);
 }
 
 void getargs(int argc, char **argv, uint8_t *foreground, uint8_t *pretend, uint8_t *loglevel, char **configfile, char **pidfile) {
