@@ -12,6 +12,8 @@
 #include "gconfig.h"
 #include "rewrite.h"
 
+#include <openssl/asn1.h>
+
 #define DEBUG_LEVEL 2
 
 #define CONFIG_MAIN SYSCONFDIR"/radsecproxy.conf"
@@ -69,6 +71,7 @@ enum rsp_mac_type {
 
 enum rsp_server_state {
     RSP_SERVER_STATE_STARTUP = 0, /* default */
+    RSP_SERVER_STATE_BLOCKING_STARTUP,
     RSP_SERVER_STATE_CONNECTED,
     RSP_SERVER_STATE_RECONNECTING,
     RSP_SERVER_STATE_FAILING
@@ -104,7 +107,7 @@ struct options {
 
 struct commonprotoopts {
     char **listenargs;
-    char *sourcearg;
+    char **sourcearg;
 };
 
 struct request {
@@ -144,16 +147,13 @@ struct clsrvconf {
     int hostaf;
     char *portsrc;
     struct list *hostports;
+    char **source;
     char *confsecret;
     uint8_t *secret;
     int secret_len;
     char *tls;
-    char *matchcertattr;
-    regex_t *certcnregex;
-    regex_t *certuriregex;
-    regex_t *certdnsregex;
-    struct in6_addr certipmatch;
-    int certipmatchaf;
+    struct list *matchcertattrs;
+    char **confmatchcertattrs;
     char *confrewritein;
     char *confrewriteout;
     char *confrewriteusername;
@@ -167,6 +167,7 @@ struct clsrvconf {
     uint8_t addttl;
     uint8_t keepalive;
     uint8_t loopprevention;
+    uint8_t blockingstartup;
     struct rewrite *rewritein;
     struct rewrite *rewriteout;
     pthread_mutex_t *lock; /* only used for updating clients so far */
