@@ -90,12 +90,21 @@ int tlsconnect(struct server *server, int timeout, char *text) {
     SSL_CTX *ctx = NULL;
     unsigned long error;
     int origflags;
+    struct addrinfo *source = NULL;
+    char *subj;
 
     debug(DBG_DBG, "tlsconnect: called from %s", text);
     pthread_mutex_lock(&server->lock);
     if (server->state == RSP_SERVER_STATE_CONNECTED)
         server->state = RSP_SERVER_STATE_RECONNECTING;
     pthread_mutex_unlock(&server->lock);
+
+    if(server->conf->source) {
+        source = resolvepassiveaddrinfo(server->conf->source, AF_UNSPEC, NULL, protodefs.socktype);
+        if(!source)
+            debug(DBG_WARN, "tlsconnect: could not resolve source address to bind for server %s, using default", server->conf->name);
+    }
+
     gettimeofday(&start, NULL);
 
     for (;;) {
@@ -109,6 +118,12 @@ int tlsconnect(struct server *server, int timeout, char *text) {
         server->ssl = NULL;
 
         wait = connect_wait(start, server->connecttime, firsttry);
+        gettimeofday(&now, NULL);
+        if (timeout && (now.tv_sec - start.tv_sec) + wait > timeout) {
+            debug(DBG_DBG, "tlsconnect: timeout");
+            if (source) freeaddrinfo(source);
+            return 0;
+        }
         debug(DBG_INFO, "Next connection attempt to %s in %lds", server->conf->name, wait);
         sleep(wait);
         firsttry = 0;
@@ -116,11 +131,12 @@ int tlsconnect(struct server *server, int timeout, char *text) {
         gettimeofday(&now, NULL);
         if (timeout && (now.tv_sec - start.tv_sec) > timeout) {
             debug(DBG_DBG, "tlsconnect: timeout");
+            if (source) freeaddrinfo(source);
             return 0;
         }
 
         debug(DBG_INFO, "tlsconnect: connecting to %s", server->conf->name);
-        if ((server->sock = connecttcphostlist(server->conf->hostports, srcres)) < 0)
+        if ((server->sock = connecttcphostlist(server->conf->hostports, source ? source : srcres)) < 0)
             continue;
         if (server->conf->keepalive)
             enable_keepalive(server->sock);
@@ -140,7 +156,7 @@ int tlsconnect(struct server *server, int timeout, char *text) {
         if (sslconnecttimeout(server->ssl, 5) <= 0) {
             while ((error = ERR_get_error()))
                 debug(DBG_ERR, "tlsconnect: SSL connect to %s failed: %s", server->conf->name, ERR_error_string(error, NULL));
-            debug(DBG_ERR, "tlsconnect: SSL connect to %s failed", server->conf->name, ERR_error_string(error, NULL));
+            debug(DBG_ERR, "tlsconnect: SSL connect to %s failed", server->conf->name);
             continue;
         }
 
@@ -148,6 +164,11 @@ int tlsconnect(struct server *server, int timeout, char *text) {
         if (!cert)
             continue;
         if (verifyconfcert(cert, server->conf)) {
+            subj = getcertsubject(cert);
+            if(subj) {
+                debug(DBG_WARN, "tlsconnect: TLS connection to %s, subject %s up", server->conf->name, subj);
+                free(subj);
+            }
             X509_free(cert);
             break;
         }
@@ -171,6 +192,7 @@ int tlsconnect(struct server *server, int timeout, char *text) {
     server->conreset = 1;
     pthread_cond_signal(&server->newrq_cond);
     pthread_mutex_unlock(&server->newrq_mutex);
+    if (source) freeaddrinfo(source);
     return 1;
 }
 
@@ -332,7 +354,7 @@ int clientradputtls(struct server *server, unsigned char *rad) {
         return 0;
     }
 
-    debug(DBG_DBG, "clientradputtls: Sent %d bytes, Radius packet of length %d to TLS peer %s", cnt, len, conf->name);
+    debug(DBG_DBG, "clientradputtls: Sent %d bytes, Radius packet of length %zu to TLS peer %s", cnt, len, conf->name);
     pthread_mutex_unlock(&server->lock);
     return 1;
 }
@@ -482,7 +504,7 @@ void *tlsservernew(void *arg) {
     unsigned long error;
     struct client *client;
     struct tls *accepted_tls = NULL;
-    char tmp[INET6_ADDRSTRLEN];
+    char tmp[INET6_ADDRSTRLEN], *subj;
 
     s = *(int *)arg;
     free(arg);
@@ -528,6 +550,12 @@ void *tlsservernew(void *arg) {
 
     while (conf) {
         if (accepted_tls == conf->tlsconf && verifyconfcert(cert, conf)) {
+            subj = getcertsubject(cert);
+            if(subj) {
+                debug(DBG_WARN, "tlsservernew: TLS connection from %s, client %s, subject %s up",
+                    addr2string((struct sockaddr *)&from,tmp, sizeof(tmp)), conf->name, subj);
+                free(subj);
+            }
             X509_free(cert);
             client = addclient(conf, 1);
             if (client) {
