@@ -18,7 +18,14 @@ struct tls {
     char *certkeypwd;
     uint8_t crlcheck;
     char **policyoids;
+    char *cipherlist;
+    char *ciphersuites;
     uint32_t cacheexpiry;
+    int tlsminversion;
+    int tlsmaxversion;
+    int dtlsminversion;
+    int dtlsmaxversion;
+    DH *dhparam;
     uint32_t tlsexpiry;
     uint32_t dtlsexpiry;
     X509_VERIFY_PARAM *vpm;
@@ -34,8 +41,10 @@ struct tls *tlsgettls(char *alt1, char *alt2);
 SSL_CTX *tlsgetctx(uint8_t type, struct tls *t);
 X509 *verifytlscert(SSL *ssl);
 int verifyconfcert(X509 *cert, struct clsrvconf *conf);
+char *getcertsubject(X509 *cert);
 int conftls_cb(struct gconffile **cf, void *arg, char *block, char *opt, char *val);
-int addmatchcertattr(struct clsrvconf *conf);
+int addmatchcertattr(struct clsrvconf *conf, const char *match);
+void freematchcertattr(struct clsrvconf *conf);
 void tlsreloadcrls();
 int sslconnecttimeout(SSL *ssl, int timeout);
 int sslaccepttimeout (SSL *ssl, int timeout);
