@@ -19,19 +19,31 @@ DIGCMD=$(command -v digaaa)
 HOSTCMD=$(command -v host)
 PRINTCMD=$(command -v printf)
 
+validate_host() {
+         echo ${@} | tr -d '\n\t\r' | grep -E '^[_0-9a-zA-Z][-._0-9a-zA-Z]*$'
+}
+
+validate_port() {
+         echo ${@} | tr -d '\n\t\r' | grep -E '^[0-9]+$'
+}
+
 dig_it() {
    ${DIGCMD} +short srv _radsec._tcp.${REALM} | sort -n -k1 |
    while read line ; do
-      set $line ; PORT=$3 ; HOST=$4 
-      $PRINTCMD "\thost ${HOST%.}:${PORT}\n"
+      set $line ; PORT=$(validate_port $3) ; HOST=$(validate_host $4)
+      if [ -n "${HOST}" ] && [ -n "${PORT}" ]; then 
+         $PRINTCMD "\thost ${HOST%.}:${PORT}\n"
+      fi
    done
 }
 
 host_it() {
    ${HOSTCMD} -t srv _radsec._tcp.${REALM} | sort -n -k5 |
    while read line ; do
-      set $line ; PORT=$7 ; HOST=$8 
-      $PRINTCMD "\thost ${HOST%.}:${PORT}\n"
+      set $line ; PORT=$(validate_port $7) ; HOST=$(validate_host $8) 
+      if [ -n "${HOST}" ] && [ -n "${PORT}" ]; then
+         $PRINTCMD "\thost ${HOST%.}:${PORT}\n"
+      fi
    done
 }
 
